@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2009-2015 Todd C. Miller <Todd.Miller@courtesan.com>
+ * Copyright (c) 2009-2017 Todd C. Miller <Todd.Miller@courtesan.com>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -194,7 +194,7 @@ main(int argc, char *argv[])
 #if defined(SUDO_DEVEL) && defined(__OpenBSD__)
     {
 	extern char *malloc_options;
-	malloc_options = "AFGJPR";
+	malloc_options = "S";
     }  
 #endif
 
@@ -222,10 +222,14 @@ main(int argc, char *argv[])
 	    /* Set the replay filter. */
 	    def_filter = false;
 	    for (cp = strtok_r(optarg, ",", &ep); cp; cp = strtok_r(NULL, ",", &ep)) {
-		if (strcmp(cp, "stdout") == 0)
+		if (strcmp(cp, "stdin") == 0)
+		    io_log_files[IOFD_STDIN].enabled = true;
+		else if (strcmp(cp, "stdout") == 0)
 		    io_log_files[IOFD_STDOUT].enabled = true;
 		else if (strcmp(cp, "stderr") == 0)
 		    io_log_files[IOFD_STDERR].enabled = true;
+		else if (strcmp(cp, "ttyin") == 0)
+		    io_log_files[IOFD_TTYIN].enabled = true;
 		else if (strcmp(cp, "ttyout") == 0)
 		    io_log_files[IOFD_TTYOUT].enabled = true;
 		else
@@ -323,12 +327,50 @@ main(int argc, char *argv[])
     /* Replay session corresponding to io_log_files[]. */
     replay_session(max_wait, decimal);
 
-    sudo_term_restore(STDIN_FILENO, 1);
+    sudo_term_restore(STDIN_FILENO, true);
 done:
     sudo_debug_exit_int(__func__, __FILE__, __LINE__, sudo_debug_subsys, exitcode);
     exit(exitcode);
 }
 
+/*
+ * Call gzread() or fread() for the I/O log file in question.
+ * Return 0 for EOF or -1 on error.
+ */
+static ssize_t
+io_log_read(int idx, char *buf, size_t nbytes)
+{
+    ssize_t nread;
+    debug_decl(io_log_read, SUDO_DEBUG_UTIL)
+
+    if (nbytes > INT_MAX) {
+	errno = EINVAL;
+	debug_return_ssize_t(-1);
+    }
+#ifdef HAVE_ZLIB_H
+    nread = gzread(io_log_files[idx].fd.g, buf, nbytes);
+#else
+    nread = (ssize_t)fread(buf, 1, nbytes, io_log_files[idx].fd.f);
+    if (nread == 0 && ferror(io_log_files[idx].fd.f))
+	nread = -1;
+#endif
+    debug_return_ssize_t(nread);
+}
+
+static char *
+io_log_gets(int idx, char *buf, size_t nbytes)
+{
+    char *str;
+    debug_decl(io_log_gets, SUDO_DEBUG_UTIL)
+
+#ifdef HAVE_ZLIB_H
+    str = gzgets(io_log_files[idx].fd.g, buf, nbytes);
+#else
+    str = fgets(buf, nbytes, io_log_files[idx].fd.f);
+#endif
+    debug_return_str(str);
+}
+
 static void
 replay_session(const double max_wait, const char *decimal)
 {
@@ -385,11 +427,7 @@ replay_session(const double max_wait, const char *decimal)
     /*
      * Read each line of the timing file, displaying the output streams.
      */
-#ifdef HAVE_ZLIB_H
-    while (gzgets(io_log_files[IOFD_TIMING].fd.g, buf, sizeof(buf)) != NULL) {
-#else
-    while (fgets(buf, sizeof(buf), io_log_files[IOFD_TIMING].fd.f) != NULL) {
-#endif
+    while (io_log_gets(IOFD_TIMING, buf, sizeof(buf)) != NULL) {
 	size_t len, nbytes, nread;
 	double seconds, to_wait;
 	struct timeval timeout;
@@ -410,11 +448,12 @@ replay_session(const double max_wait, const char *decimal)
 	timeout.tv_usec = (to_wait - timeout.tv_sec) * 1000000.0;
 
 	/* Run event event loop to delay and get keyboard input. */
-	sudo_ev_add(evbase, input_ev, &timeout, false);
+	if (sudo_ev_add(evbase, input_ev, &timeout, false) == -1)
+	    sudo_fatal(NULL);
 	sudo_ev_loop(evbase, 0);
 
 	/* Even if we are not replaying, we still have to delay. */
-	if (io_log_files[idx].fd.v == NULL)
+	if (idx >= IOFD_MAX || io_log_files[idx].fd.v == NULL)
 	    continue;
 
 	/* Check whether we need to convert newline to CR LF pairs. */
@@ -428,11 +467,18 @@ replay_session(const double max_wait, const char *decimal)
 		len = sizeof(buf);
 	    else
 		len = nbytes;
-#ifdef HAVE_ZLIB_H
-	    nread = gzread(io_log_files[idx].fd.g, buf, len);
-#else
-	    nread = fread(buf, 1, len, io_log_files[idx].fd.f);
-#endif
+	    nread = io_log_read(idx, buf, len);
+	    if (nread <= 0) {
+		if (nread == 0) {
+		    sudo_debug_printf(SUDO_DEBUG_ERROR|SUDO_DEBUG_LINENO,
+			"%s: premature EOF, expected %zu bytes",
+			io_log_files[idx].suffix, nbytes);
+		} else {
+		    sudo_debug_printf(SUDO_DEBUG_ERROR|SUDO_DEBUG_ERRNO|SUDO_DEBUG_LINENO,
+			"%s: read error", io_log_files[idx].suffix);
+		}
+		break;
+	    }
 	    nbytes -= nread;
 
 	    /* Convert newline to carriage return + linefeed if needed. */
@@ -508,12 +554,16 @@ replay_session(const double max_wait, const char *decimal)
 
 	    /* Run event event loop to write output. */
 	    /* XXX - should use a single event loop with a circular buffer. */
-	    sudo_ev_add(evbase, output_ev, NULL, false);
+	    if (sudo_ev_add(evbase, output_ev, NULL, false) == -1)
+		sudo_fatal(NULL);
 	    sudo_ev_loop(evbase, 0);
 	}
-	if (iov != &iovb)
-	    free(iov);
     }
+    if (iov != &iovb)
+	free(iov);
+    sudo_ev_base_free(evbase);
+    sudo_ev_free(input_ev);
+    sudo_ev_free(output_ev);
     debug_return;
 }
 
@@ -575,7 +625,8 @@ write_output(int fd, int what, void *v)
     }
 
     /* Reschedule event to write remainder. */
-    sudo_ev_add(sudo_ev_get_base(wc->wevent), wc->wevent, NULL, false);
+    if (sudo_ev_add(NULL, wc->wevent, NULL, false) == -1)
+	sudo_fatal(NULL);
     debug_return;
 }
 
@@ -606,7 +657,7 @@ parse_expr(struct search_node_list *head, char *argv[], bool sub_expr)
 		goto bad;
 	    not = true;
 	    continue;
-	case 'c': /* command */
+	case 'c': /* cwd or command */
 	    if (av[0][1] == '\0')
 		sudo_fatalx(U_("ambiguous expression \"%s\""), *av);
 	    if (strncmp(*av, "cwd", strlen(*av)) == 0)
@@ -657,8 +708,8 @@ parse_expr(struct search_node_list *head, char *argv[], bool sub_expr)
 	    if (!sub_expr)
 		sudo_fatalx(U_("unmatched ')' in expression"));
 	    debug_return_int(av - argv + 1);
-	bad:
 	default:
+	bad:
 	    sudo_fatalx(U_("unknown search term \"%s\""), *av);
 	    /* NOTREACHED */
 	}
@@ -900,7 +951,7 @@ list_session(char *logfile, regex_t *re, const char *user, const char *tty)
     char idbuf[7], *idstr, *cp;
     const char *timestr;
     struct log_info *li;
-    int rval = -1;
+    int ret = -1;
     debug_decl(list_session, SUDO_DEBUG_UTIL)
 
     if ((li = parse_logfile(logfile)) == NULL)
@@ -935,11 +986,11 @@ list_session(char *logfile, regex_t *re, const char *user, const char *tty)
 	printf("GROUP=%s ; ", li->runas_group);
     printf("TSID=%s ; COMMAND=%s\n", idstr, li->cmd);
 
-    rval = 0;
+    ret = 0;
 
 done:
     free_log_info(li);
-    debug_return_int(rval);
+    debug_return_int(ret);
 }
 
 static int
@@ -1071,7 +1122,6 @@ static void
 check_input(int fd, int what, void *v)
 {
     struct sudo_event *ev = v;
-    struct sudo_event_base *evbase = sudo_ev_get_base(ev);
     struct timeval tv, *timeout = NULL;
     static bool paused = 0;
     char ch;
@@ -1086,7 +1136,7 @@ check_input(int fd, int what, void *v)
 	case 0:
 	    /* Ignore EOF. */
 	    break;
-	case 1:
+	default:
 	    if (paused) {
 		/* Any key will unpause, event is finished. */
 		/* XXX - pause time could be less than timeout */
@@ -1119,7 +1169,8 @@ check_input(int fd, int what, void *v)
 	    timeout = &tv;
 	}
 	/* Re-enable event. */
-	sudo_ev_add(evbase, ev, timeout, false);
+	if (sudo_ev_add(NULL, ev, timeout, false) == -1)
+	    sudo_fatal(NULL);
     }
     debug_return;
 }
@@ -1225,7 +1276,7 @@ help(void)
 static void
 sudoreplay_cleanup(void)
 {
-    sudo_term_restore(STDIN_FILENO, 0);
+    sudo_term_restore(STDIN_FILENO, false);
 }
 
 /*
@@ -1235,6 +1286,6 @@ sudoreplay_cleanup(void)
 static void
 sudoreplay_handler(int signo)
 {
-    sudo_term_restore(STDIN_FILENO, 0);
+    sudo_term_restore(STDIN_FILENO, false);
     kill(getpid(), signo);
 }
