@@ -1,5 +1,6 @@
 /*
- * Copyright (c) 1999, 2009-2011 Todd C. Miller <Todd.Miller@courtesan.com>
+ * Copyright (c) 1999, 2009-2011, 2013-2015, 2017
+ *	Todd C. Miller <Todd.Miller@courtesan.com>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -18,50 +19,28 @@
 
 #include <sys/types.h>
 #include <stdio.h>
-#ifdef STDC_HEADERS
-# include <stdlib.h>
-# include <stddef.h>
-#else
-# ifdef HAVE_STDLIB_H
-#  include <stdlib.h>
-# endif
-#endif /* STDC_HEADERS */
+#include <stdlib.h>
 #include <time.h>
 
-#include "missing.h"
-
-char *get_timestr(time_t, int);
+#include "sudo_compat.h"
+#include "sudo_debug.h"
+#include "parse.h"
 
 /*
- * Return an ascii string with the current date + time
- * Uses strftime() if available, else falls back to ctime().
+ * Return a static buffer with the current date + time.
  */
 char *
 get_timestr(time_t tstamp, int log_year)
 {
-    char *s;
-#ifdef HAVE_STRFTIME
     static char buf[128];
     struct tm *timeptr;
 
-    timeptr = localtime(&tstamp);
-    if (log_year)
-	s = "%h %e %T %Y";
-    else
-	s = "%h %e %T";
-
-    /* strftime() does not guarantee to NUL-terminate so we must check. */
-    buf[sizeof(buf) - 1] = '\0';
-    if (strftime(buf, sizeof(buf), s, timeptr) && buf[sizeof(buf) - 1] == '\0')
-	return buf;
-
-#endif /* HAVE_STRFTIME */
-
-    s = ctime(&tstamp) + 4;		/* skip day of the week */
-    if (log_year)
-	s[20] = '\0';			/* avoid the newline */
-    else
-	s[15] = '\0';			/* don't care about year */
-
-    return s;
+    if ((timeptr = localtime(&tstamp)) != NULL) {
+	/* strftime() does not guarantee to NUL-terminate so we must check. */
+	buf[sizeof(buf) - 1] = '\0';
+	if (strftime(buf, sizeof(buf), log_year ? "%h %e %T %Y" : "%h %e %T",
+	    timeptr) != 0 && buf[sizeof(buf) - 1] == '\0')
+	    return buf;
+    }
+    return NULL;
 }
