@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1993-1996, 1998-2012 Todd C. Miller <Todd.Miller@courtesan.com>
+ * Copyright (c) 1993-1996, 1998-2017 Todd C. Miller <Todd.Miller@courtesan.com>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -21,37 +21,29 @@
 #include <config.h>
 
 #include <sys/types.h>
-#include <sys/param.h>
 
 #include <stdio.h>
-#ifdef STDC_HEADERS
-# include <stdlib.h>
-# include <stddef.h>
-#else
-# ifdef HAVE_STDLIB_H
-#  include <stdlib.h>
-# endif
-#endif /* STDC_HEADERS */
+#include <stdlib.h>
 #ifdef HAVE_STRING_H
 # include <string.h>
 #endif /* HAVE_STRING_H */
 #ifdef HAVE_STRINGS_H
 # include <strings.h>
 #endif /* HAVE_STRINGS_H */
-#ifdef HAVE_UNISTD_H
-# include <unistd.h>
-#endif /* HAVE_UNISTD_H */
+#include <unistd.h>
 #include <ctype.h>
 #include <grp.h>
 #include <pwd.h>
 
 #include <sudo_usage.h>
 #include "sudo.h"
-#include "lbuf.h"
+#include "sudo_lbuf.h"
 
-/* For getopt(3) */
-extern char *optarg;
-extern int optind;
+#ifdef HAVE_GETOPT_LONG
+# include <getopt.h>
+# else
+# include "compat/getopt.h"
+#endif /* HAVE_GETOPT_LONG */
 
 int tgetpass_flags;
 
@@ -64,80 +56,140 @@ static void usage_excl(int);
 /*
  * Mapping of command line flags to name/value settings.
  */
-static struct sudo_settings {
-    const char *name;
-    const char *value;
-} sudo_settings[] = {
+static struct sudo_settings sudo_settings[] = {
 #define ARG_BSDAUTH_TYPE 0
     { "bsdauth_type" },
 #define ARG_LOGIN_CLASS 1
     { "login_class" },
-#define ARG_DEBUG_FLAGS 2
-    { "debug_flags" },
-#define ARG_PRESERVE_ENVIRONMENT 3
+#define ARG_PRESERVE_ENVIRONMENT 2
     { "preserve_environment" },
-#define ARG_RUNAS_GROUP 4
+#define ARG_RUNAS_GROUP 3
     { "runas_group" },
-#define ARG_SET_HOME 5
+#define ARG_SET_HOME 4
     { "set_home" },
-#define ARG_USER_SHELL 6
+#define ARG_USER_SHELL 5
     { "run_shell" },
-#define ARG_LOGIN_SHELL 7
+#define ARG_LOGIN_SHELL 6
     { "login_shell" },
-#define ARG_IGNORE_TICKET 8
+#define ARG_IGNORE_TICKET 7
     { "ignore_ticket" },
-#define ARG_PROMPT 9
+#define ARG_PROMPT 8
     { "prompt" },
-#define ARG_SELINUX_ROLE 10
+#define ARG_SELINUX_ROLE 9
     { "selinux_role" },
-#define ARG_SELINUX_TYPE 11
+#define ARG_SELINUX_TYPE 10
     { "selinux_type" },
-#define ARG_RUNAS_USER 12
+#define ARG_RUNAS_USER 11
     { "runas_user" },
-#define ARG_PROGNAME 13
+#define ARG_PROGNAME 12
     { "progname" },
-#define ARG_IMPLIED_SHELL 14
+#define ARG_IMPLIED_SHELL 13
     { "implied_shell" },
-#define ARG_PRESERVE_GROUPS 15
+#define ARG_PRESERVE_GROUPS 14
     { "preserve_groups" },
-#define ARG_NONINTERACTIVE 16
+#define ARG_NONINTERACTIVE 15
     { "noninteractive" },
-#define ARG_SUDOEDIT 17
+#define ARG_SUDOEDIT 16
     { "sudoedit" },
-#define ARG_CLOSEFROM 18
+#define ARG_CLOSEFROM 17
     { "closefrom" },
-#define ARG_NET_ADDRS 19
+#define ARG_NET_ADDRS 18
     { "network_addrs" },
-#define NUM_SETTINGS 20
+#define ARG_MAX_GROUPS 19
+    { "max_groups" },
+#define ARG_PLUGIN_DIR 20
+    { "plugin_dir" },
+#define ARG_REMOTE_HOST 21
+    { "remote_host" },
+#define ARG_TIMEOUT 22
+    { "timeout" },
+#define NUM_SETTINGS 23
     { NULL }
 };
 
+/*
+ * Default flags allowed when running a command.
+ */
+#define DEFAULT_VALID_FLAGS	(MODE_BACKGROUND|MODE_PRESERVE_ENV|MODE_RESET_HOME|MODE_LOGIN_SHELL|MODE_NONINTERACTIVE|MODE_SHELL)
+
+/* Option number for the --host long option due to ambiguity of the -h flag. */
+#define OPT_HOSTNAME	256
+
+/*
+ * Available command line options, both short and long.
+ * Note that we must disable arg permutation to support setting environment
+ * variables and to better support the optional arg of the -h flag.
+ */
+static const char short_opts[] =  "+Aa:bC:c:D:Eeg:Hh::iKklnPp:r:SsT:t:U:u:Vv";
+static struct option long_opts[] = {
+    { "askpass",	no_argument,		NULL,	'A' },
+    { "auth-type",	required_argument,	NULL,	'a' },
+    { "background",	no_argument,		NULL,	'b' },
+    { "close-from",	required_argument,	NULL,	'C' },
+    { "login-class",	required_argument,	NULL,	'c' },
+    { "preserve-env",	no_argument,		NULL,	'E' },
+    { "edit",		no_argument,		NULL,	'e' },
+    { "group",		required_argument,	NULL,	'g' },
+    { "set-home",	no_argument,		NULL,	'H' },
+    { "help",		no_argument,		NULL,	'h' },
+    { "host",		required_argument,	NULL,	OPT_HOSTNAME },
+    { "login",		no_argument,		NULL,	'i' },
+    { "remove-timestamp", no_argument,		NULL,	'K' },
+    { "reset-timestamp", no_argument,		NULL,	'k' },
+    { "list",		no_argument,		NULL,	'l' },
+    { "non-interactive", no_argument,		NULL,	'n' },
+    { "preserve-groups", no_argument,		NULL,	'P' },
+    { "prompt",		required_argument,	NULL,	'p' },
+    { "role",		required_argument,	NULL,	'r' },
+    { "stdin",		no_argument,		NULL,	'S' },
+    { "shell",		no_argument,		NULL,	's' },
+    { "type",		required_argument,	NULL,	't' },
+    { "command-timeout",required_argument,	NULL,	'T' },
+    { "other-user",	required_argument,	NULL,	'U' },
+    { "user",		required_argument,	NULL,	'u' },
+    { "version",	no_argument,		NULL,	'V' },
+    { "validate",	no_argument,		NULL,	'v' },
+    { NULL,		no_argument,		NULL,	'\0' },
+};
+
 /*
  * Command line argument parsing.
  * Sets nargc and nargv which corresponds to the argc/argv we'll use
  * for the command to be run (if we are running one).
  */
 int
-parse_args(int argc, char **argv, int *nargc, char ***nargv, char ***settingsp,
-    char ***env_addp)
+parse_args(int argc, char **argv, int *nargc, char ***nargv,
+    struct sudo_settings **settingsp, char ***env_addp)
 {
     int mode = 0;		/* what mode is sudo to be run in? */
     int flags = 0;		/* mode flags */
-    int valid_flags, ch;
-    int i, j;
-    char *cp, **env_add, **settings;
-    const char *debug_flags;
+    int valid_flags = DEFAULT_VALID_FLAGS;
+    int ch, i;
+    char *cp, **env_add;
+    const char *runas_user = NULL;
+    const char *runas_group = NULL;
+    const char *progname;
+    int proglen;
     int nenv = 0;
     int env_size = 32;
     debug_decl(parse_args, SUDO_DEBUG_ARGS)
 
-    env_add = emalloc2(env_size, sizeof(char *));
+    /* Is someone trying something funny? */
+    if (argc <= 0)
+	usage(1);
 
-    /* Pass progname to plugin so it can call setprogname() */
-    sudo_settings[ARG_PROGNAME].value = getprogname();
+    env_add = reallocarray(NULL, env_size, sizeof(char *));
+    if (env_add == NULL)
+	sudo_fatalx(U_("%s: %s"), __func__, U_("unable to allocate memory"));
+
+    /* Pass progname to plugin so it can call initprogname() */
+    progname = getprogname();
+    sudo_settings[ARG_PROGNAME].value = progname;
 
     /* First, check to see if we were invoked as "sudoedit". */
-    if (strcmp(getprogname(), "sudoedit") == 0) {
+    proglen = strlen(progname);
+    if (proglen > 4 && strcmp(progname + proglen - 4, "edit") == 0) {
+	progname = "sudoedit";
 	mode = MODE_EDIT;
 	sudo_settings[ARG_SUDOEDIT].value = "true";
     }
@@ -146,10 +198,17 @@ parse_args(int argc, char **argv, int *nargc, char ***nargv, char ***settingsp,
     if (get_net_ifs(&cp) > 0)
 	sudo_settings[ARG_NET_ADDRS].value = cp;
 
-    /* Set debug file and flags from sudo.conf. */
-    debug_flags = sudo_conf_debug_flags();
-    if (debug_flags != NULL)
-	sudo_settings[ARG_DEBUG_FLAGS].value = debug_flags;
+    /* Set max_groups from sudo.conf. */
+    i = sudo_conf_max_groups();
+    if (i != -1) {
+	if (asprintf(&cp, "%d", i) == -1)
+	    sudo_fatalx(U_("%s: %s"), __func__, U_("unable to allocate memory"));
+	sudo_settings[ARG_MAX_GROUPS].value = cp;
+    }
+
+    /* Returns true if the last option string was "-h" */
+#define got_host_flag	(optind > 1 && argv[optind - 1][0] == '-' && \
+	    argv[optind - 1][1] == 'h' && argv[optind - 1][2] == '\0')
 
     /* Returns true if the last option string was "--" */
 #define got_end_of_args	(optind > 1 && argv[optind - 1][0] == '-' && \
@@ -159,17 +218,13 @@ parse_args(int argc, char **argv, int *nargc, char ***nargv, char ***settingsp,
 #define is_envar (optind < argc && argv[optind][0] != '/' && \
 	    strchr(argv[optind], '=') != NULL)
 
-    /* Flags allowed when running a command */
-    valid_flags = MODE_BACKGROUND|MODE_PRESERVE_ENV|MODE_RESET_HOME|
-		  MODE_LOGIN_SHELL|MODE_NONINTERACTIVE|MODE_SHELL;
     /* XXX - should fill in settings at the end to avoid dupes */
     for (;;) {
 	/*
-	 * We disable arg permutation for GNU getopt().
 	 * Some trickiness is required to allow environment variables
 	 * to be interspersed with command line options.
 	 */
-	if ((ch = getopt(argc, argv, "+Aa:bC:c:D:Eeg:HhiKklnPp:r:Sst:U:u:Vv")) != -1) {
+	if ((ch = getopt_long(argc, argv, short_opts, long_opts, NULL)) != -1) {
 	    switch (ch) {
 		case 'A':
 		    SET(tgetpass_flags, TGP_ASKPASS);
@@ -183,8 +238,8 @@ parse_args(int argc, char **argv, int *nargc, char ***nargv, char ***settingsp,
 		    SET(flags, MODE_BACKGROUND);
 		    break;
 		case 'C':
-		    if (atoi(optarg) < 3) {
-			warningx(_("the argument to -C must be a number greater than or equal to 3"));
+		    if (strtonum(optarg, 3, INT_MAX, NULL) == 0) {
+			sudo_warnx(_("the argument to -C must be a number greater than or equal to 3"));
 			usage(1);
 		    }
 		    sudo_settings[ARG_CLOSEFROM].value = optarg;
@@ -215,12 +270,28 @@ parse_args(int argc, char **argv, int *nargc, char ***nargv, char ***settingsp,
 		    sudo_settings[ARG_SET_HOME].value = "true";
 		    break;
 		case 'h':
-		    if (mode && mode != MODE_HELP) {
-			if (strcmp(getprogname(), "sudoedit") != 0)
-			    usage_excl(1);
+		    if (optarg == NULL) {
+			/*
+			 * Optional args support -hhostname, not -h hostname.
+			 * If we see a non-option after the -h flag, treat as
+			 * remote host and bump optind to skip over it.
+			 */
+			if (got_host_flag && !is_envar &&
+			    argv[optind] != NULL && argv[optind][0] != '-') {
+			    sudo_settings[ARG_REMOTE_HOST].value = argv[optind++];
+			    continue;
+			}
+			if (mode && mode != MODE_HELP) {
+			    if (strcmp(progname, "sudoedit") != 0)
+				usage_excl(1);
+			}
+			mode = MODE_HELP;
+			valid_flags = 0;
+			break;
 		    }
-		    mode = MODE_HELP;
-		    valid_flags = 0;
+		    /* FALLTHROUGH */
+		case OPT_HOSTNAME:
+		    sudo_settings[ARG_REMOTE_HOST].value = optarg;
 		    break;
 		case 'i':
 		    sudo_settings[ARG_LOGIN_SHELL].value = "true";
@@ -264,6 +335,9 @@ parse_args(int argc, char **argv, int *nargc, char ***nargv, char ***settingsp,
 		    sudo_settings[ARG_SELINUX_TYPE].value = optarg;
 		    break;
 #endif
+		case 'T':
+		    sudo_settings[ARG_TIMEOUT].value = optarg;
+		    break;
 		case 'S':
 		    SET(tgetpass_flags, TGP_STDIN);
 		    break;
@@ -272,8 +346,6 @@ parse_args(int argc, char **argv, int *nargc, char ***nargv, char ***settingsp,
 		    SET(flags, MODE_SHELL);
 		    break;
 		case 'U':
-		    if ((getpwnam(optarg)) == NULL)
-			errorx(1, _("unknown user: %s"), optarg);
 		    list_user = optarg;
 		    break;
 		case 'u':
@@ -297,8 +369,13 @@ parse_args(int argc, char **argv, int *nargc, char ***nargv, char ***settingsp,
 	    }
 	} else if (!got_end_of_args && is_envar) {
 	    if (nenv == env_size - 2) {
+		char **tmp;
+
+		tmp = reallocarray(env_add, env_size, 2 * sizeof(char *));
+		if (tmp == NULL)
+		    sudo_fatalx(U_("%s: %s"), __func__, U_("unable to allocate memory"));
+		env_add = tmp;
 		env_size *= 2;
-		env_add = erealloc3(env_add, env_size, sizeof(char *));
 	    }
 	    env_add[nenv++] = argv[optind];
 
@@ -332,11 +409,11 @@ parse_args(int argc, char **argv, int *nargc, char ***nargv, char ***settingsp,
 
     if (ISSET(flags, MODE_LOGIN_SHELL)) {
 	if (ISSET(flags, MODE_SHELL)) {
-	    warningx(_("you may not specify both the `-i' and `-s' options"));
+	    sudo_warnx(U_("you may not specify both the `-i' and `-s' options"));
 	    usage(1);
 	}
 	if (ISSET(flags, MODE_PRESERVE_ENV)) {
-	    warningx(_("you may not specify both the `-i' and `-E' options"));
+	    sudo_warnx(U_("you may not specify both the `-i' and `-E' options"));
 	    usage(1);
 	}
 	SET(flags, MODE_SHELL);
@@ -346,9 +423,9 @@ parse_args(int argc, char **argv, int *nargc, char ***nargv, char ***settingsp,
     if (mode == MODE_EDIT &&
        (ISSET(flags, MODE_PRESERVE_ENV) || env_add[0] != NULL)) {
 	if (ISSET(mode, MODE_PRESERVE_ENV))
-	    warningx(_("the `-E' option is not valid in edit mode"));
+	    sudo_warnx(U_("the `-E' option is not valid in edit mode"));
 	if (env_add[0] != NULL)
-	    warningx(_("you may not specify environment variables in edit mode"));
+	    sudo_warnx(U_("you may not specify environment variables in edit mode"));
 	usage(1);
     }
     if ((runas_user != NULL || runas_group != NULL) &&
@@ -356,11 +433,11 @@ parse_args(int argc, char **argv, int *nargc, char ***nargv, char ***settingsp,
 	usage(1);
     }
     if (list_user != NULL && mode != MODE_LIST && mode != MODE_CHECK) {
-	warningx(_("the `-U' option may only be used with the `-l' option"));
+	sudo_warnx(U_("the `-U' option may only be used with the `-l' option"));
 	usage(1);
     }
     if (ISSET(tgetpass_flags, TGP_STDIN) && ISSET(tgetpass_flags, TGP_ASKPASS)) {
-	warningx(_("the `-A' and `-S' options may not be used together"));
+	sudo_warnx(U_("the `-A' and `-S' options may not be used together"));
 	usage(1);
     }
     if ((argc == 0 && mode == MODE_EDIT) ||
@@ -370,6 +447,9 @@ parse_args(int argc, char **argv, int *nargc, char ***nargv, char ***settingsp,
 	SET(flags, (MODE_IMPLIED_SHELL | MODE_SHELL));
 	sudo_settings[ARG_IMPLIED_SHELL].value = "true";
     }
+#ifdef ENABLE_SUDO_PLUGIN_API
+    sudo_settings[ARG_PLUGIN_DIR].value = sudo_conf_plugin_dir_path();
+#endif
 
     if (mode == MODE_HELP)
 	help();
@@ -387,11 +467,16 @@ parse_args(int argc, char **argv, int *nargc, char ***nargv, char ***settingsp,
 	    size_t cmnd_size = (size_t) (argv[argc - 1] - argv[0]) +
 		strlen(argv[argc - 1]) + 1;
 
-	    cmnd = dst = emalloc2(cmnd_size, 2);
+	    cmnd = dst = reallocarray(NULL, cmnd_size, 2);
+	    if (cmnd == NULL)
+		sudo_fatalx(U_("%s: %s"), __func__, U_("unable to allocate memory"));
+	    if (!gc_add(GC_PTR, cmnd))
+		exit(1);
+
 	    for (av = argv; *av != NULL; av++) {
 		for (src = *av; *src != '\0'; src++) {
 		    /* quote potential meta characters */
-		    if (!isalnum((unsigned char)*src) && *src != '_' && *src != '-')
+		    if (!isalnum((unsigned char)*src) && *src != '_' && *src != '-' && *src != '$')
 			*dst++ = '\\';
 		    *dst++ = *src;
 		}
@@ -404,7 +489,12 @@ parse_args(int argc, char **argv, int *nargc, char ***nargv, char ***settingsp,
 	    ac += 2; /* -c cmnd */
 	}
 
-	av = emalloc2(ac + 1, sizeof(char *));
+	av = reallocarray(NULL, ac + 1, sizeof(char *));
+	if (av == NULL)
+	    sudo_fatalx(U_("%s: %s"), __func__, U_("unable to allocate memory"));
+	if (!gc_add(GC_PTR, av))
+	    exit(1);
+
 	av[0] = (char *)user_details.shell; /* plugin may override shell */
 	if (cmnd != NULL) {
 	    av[1] = "-c";
@@ -416,23 +506,6 @@ parse_args(int argc, char **argv, int *nargc, char ***nargv, char ***settingsp,
 	argc = ac;
     }
 
-    /*
-     * Format setting_pairs into settings array.
-     */
-    settings = emalloc2(NUM_SETTINGS + 1, sizeof(char *));
-    for (i = 0, j = 0; i < NUM_SETTINGS; i++) {
-	if (sudo_settings[i].value) {
-	    sudo_debug_printf(SUDO_DEBUG_INFO, "settings: %s=%s",
-		sudo_settings[i].name, sudo_settings[i].value);
-	    settings[j] = fmt_string(sudo_settings[i].name,
-		sudo_settings[i].value);
-	    if (settings[j] == NULL)
-		errorx(1, _("unable to allocate memory"));
-	    j++;
-	}
-    }
-    settings[j] = NULL;
-
     if (mode == MODE_EDIT) {
 #if defined(HAVE_SETRESUID) || defined(HAVE_SETREUID) || defined(HAVE_SETEUID)
 	/* Must have the command in argv[0]. */
@@ -440,11 +513,11 @@ parse_args(int argc, char **argv, int *nargc, char ***nargv, char ***settingsp,
 	argv--;
 	argv[0] = "sudoedit";
 #else
-	errorx(1, _("sudoedit is not supported on this platform"));
+	sudo_fatalx(U_("sudoedit is not supported on this platform"));
 #endif
     }
 
-    *settingsp = settings;
+    *settingsp = sudo_settings;
     *env_addp = env_add;
     *nargc = argc;
     *nargv = argv;
@@ -470,7 +543,7 @@ usage_out(const char *buf)
 void
 usage(int fatal)
 {
-    struct lbuf lbuf;
+    struct sudo_lbuf lbuf;
     char *uvec[6];
     int i, ulen;
 
@@ -494,13 +567,13 @@ usage(int fatal)
      * tty width.
      */
     ulen = (int)strlen(getprogname()) + 8;
-    lbuf_init(&lbuf, fatal ? usage_err : usage_out, ulen, NULL,
+    sudo_lbuf_init(&lbuf, fatal ? usage_err : usage_out, ulen, NULL,
 	user_details.ts_cols);
     for (i = 0; uvec[i] != NULL; i++) {
-	lbuf_append(&lbuf, "usage: %s%s", getprogname(), uvec[i]);
-	lbuf_print(&lbuf);
+	sudo_lbuf_append(&lbuf, "usage: %s%s", getprogname(), uvec[i]);
+	sudo_lbuf_print(&lbuf);
     }
-    lbuf_destroy(&lbuf);
+    sudo_lbuf_destroy(&lbuf);
     if (fatal)
 	exit(1);
 }
@@ -513,90 +586,94 @@ usage_excl(int fatal)
 {
     debug_decl(usage_excl, SUDO_DEBUG_ARGS)
 
-    warningx(_("Only one of the -e, -h, -i, -K, -l, -s, -v or -V options may be specified"));
+    sudo_warnx(U_("Only one of the -e, -h, -i, -K, -l, -s, -v or -V options may be specified"));
     usage(fatal);
 }
 
 static void
 help(void)
 {
-    struct lbuf lbuf;
-    int indent = 16;
+    struct sudo_lbuf lbuf;
+    const int indent = 32;
     const char *pname = getprogname();
     debug_decl(help, SUDO_DEBUG_ARGS)
 
-    lbuf_init(&lbuf, usage_out, indent, NULL, user_details.ts_cols);
+    sudo_lbuf_init(&lbuf, usage_out, indent, NULL, user_details.ts_cols);
     if (strcmp(pname, "sudoedit") == 0)
-	lbuf_append(&lbuf, _("%s - edit files as another user\n\n"), pname);
+	sudo_lbuf_append(&lbuf, _("%s - edit files as another user\n\n"), pname);
     else
-	lbuf_append(&lbuf, _("%s - execute a command as another user\n\n"), pname);
-    lbuf_print(&lbuf);
+	sudo_lbuf_append(&lbuf, _("%s - execute a command as another user\n\n"), pname);
+    sudo_lbuf_print(&lbuf);
 
     usage(0);
 
-    lbuf_append(&lbuf, _("\nOptions:\n"));
-    lbuf_append(&lbuf, "  -A            %s",
-	_("use helper program for password prompting\n"));
+    sudo_lbuf_append(&lbuf, _("\nOptions:\n"));
+    sudo_lbuf_append(&lbuf, "  -A, --askpass                 %s\n",
+	_("use a helper program for password prompting"));
 #ifdef HAVE_BSD_AUTH_H
-    lbuf_append(&lbuf, "  -a type       %s",
-	_("use specified BSD authentication type\n"));
+    sudo_lbuf_append(&lbuf, "  -a, --auth-type=type          %s\n",
+	_("use specified BSD authentication type"));
 #endif
-    lbuf_append(&lbuf, "  -b            %s",
-	_("run command in the background\n"));
-    lbuf_append(&lbuf, "  -C fd         %s",
-	_("close all file descriptors >= fd\n"));
+    sudo_lbuf_append(&lbuf, "  -b, --background              %s\n",
+	_("run command in the background"));
+    sudo_lbuf_append(&lbuf, "  -C, --close-from=num          %s\n",
+	_("close all file descriptors >= num"));
 #ifdef HAVE_LOGIN_CAP_H
-    lbuf_append(&lbuf, "  -c class      %s",
-	_("run command with specified login class\n"));
+    sudo_lbuf_append(&lbuf, "  -c, --login-class=class       %s\n",
+	_("run command with the specified BSD login class"));
 #endif
-    lbuf_append(&lbuf, "  -E            %s",
-	_("preserve user environment when executing command\n"));
-    lbuf_append(&lbuf, "  -e            %s",
-	_("edit files instead of running a command\n"));
-    lbuf_append(&lbuf, "  -g group      %s",
-	_("execute command as the specified group\n"));
-    lbuf_append(&lbuf, "  -H            %s",
-	_("set HOME variable to target user's home dir.\n"));
-    lbuf_append(&lbuf, "  -h            %s",
-	_("display help message and exit\n"));
-    lbuf_append(&lbuf, "  -i [command]  %s",
-	_("run a login shell as target user\n"));
-    lbuf_append(&lbuf, "  -K            %s",
-	_("remove timestamp file completely\n"));
-    lbuf_append(&lbuf, "  -k            %s",
-	_("invalidate timestamp file\n"));
-    lbuf_append(&lbuf, "  -l[l] command %s",
-	_("list user's available commands\n"));
-    lbuf_append(&lbuf, "  -n            %s",
-	_("non-interactive mode, will not prompt user\n"));
-    lbuf_append(&lbuf, "  -P            %s",
-	_("preserve group vector instead of setting to target's\n"));
-    lbuf_append(&lbuf, "  -p prompt     %s",
-	_("use specified password prompt\n"));
+    sudo_lbuf_append(&lbuf, "  -E, --preserve-env            %s\n",
+	_("preserve user environment when running command"));
+    sudo_lbuf_append(&lbuf, "  -e, --edit                    %s\n",
+	_("edit files instead of running a command"));
+    sudo_lbuf_append(&lbuf, "  -g, --group=group             %s\n",
+	_("run command as the specified group name or ID"));
+    sudo_lbuf_append(&lbuf, "  -H, --set-home                %s\n",
+	_("set HOME variable to target user's home dir"));
+    sudo_lbuf_append(&lbuf, "  -h, --help                    %s\n",
+	_("display help message and exit"));
+    sudo_lbuf_append(&lbuf, "  -h, --host=host               %s\n",
+	_("run command on host (if supported by plugin)"));
+    sudo_lbuf_append(&lbuf, "  -i, --login                   %s\n",
+	_("run login shell as the target user; a command may also be specified"));
+    sudo_lbuf_append(&lbuf, "  -K, --remove-timestamp        %s\n",
+	_("remove timestamp file completely"));
+    sudo_lbuf_append(&lbuf, "  -k, --reset-timestamp         %s\n",
+	_("invalidate timestamp file"));
+    sudo_lbuf_append(&lbuf, "  -l, --list                    %s\n",
+	_("list user's privileges or check a specific command; use twice for longer format"));
+    sudo_lbuf_append(&lbuf, "  -n, --non-interactive         %s\n",
+	_("non-interactive mode, no prompts are used"));
+    sudo_lbuf_append(&lbuf, "  -P, --preserve-groups         %s\n",
+	_("preserve group vector instead of setting to target's"));
+    sudo_lbuf_append(&lbuf, "  -p, --prompt=prompt           %s\n",
+	_("use the specified password prompt"));
 #ifdef HAVE_SELINUX
-    lbuf_append(&lbuf, "  -r role       %s",
-	_("create SELinux security context with specified role\n"));
+    sudo_lbuf_append(&lbuf, "  -r, --role=role               %s\n",
+	_("create SELinux security context with specified role"));
 #endif
-    lbuf_append(&lbuf, "  -S            %s",
-	_("read password from standard input\n"));
-    lbuf_append(&lbuf,
-	"  -s [command]  %s", _("run a shell as target user\n"));
+    sudo_lbuf_append(&lbuf, "  -S, --stdin                   %s\n",
+	_("read password from standard input"));
+    sudo_lbuf_append(&lbuf, "  -s, --shell                   %s\n",
+	_("run shell as the target user; a command may also be specified"));
 #ifdef HAVE_SELINUX
-    lbuf_append(&lbuf, "  -t type       %s",
-	_("create SELinux security context with specified role\n"));
+    sudo_lbuf_append(&lbuf, "  -t, --type=type               %s\n",
+	_("create SELinux security context with specified type"));
 #endif
-    lbuf_append(&lbuf, "  -U user       %s",
-	_("when listing, list specified user's privileges\n"));
-    lbuf_append(&lbuf, "  -u user       %s",
-	_("run command (or edit file) as specified user\n"));
-    lbuf_append(&lbuf, "  -V            %s",
-	_("display version information and exit\n"));
-    lbuf_append(&lbuf, "  -v            %s",
-	_("update user's timestamp without running a command\n"));
-    lbuf_append(&lbuf, "  --            %s",
-	_("stop processing command line arguments\n"));
-    lbuf_print(&lbuf);
-    lbuf_destroy(&lbuf);
+    sudo_lbuf_append(&lbuf, "  -T, --command-timeout=timeout %s\n",
+	_("terminate command after the specified time limit"));
+    sudo_lbuf_append(&lbuf, "  -U, --other-user=user         %s\n",
+	_("in list mode, display privileges for user"));
+    sudo_lbuf_append(&lbuf, "  -u, --user=user               %s\n",
+	_("run command (or edit file) as specified user name or ID"));
+    sudo_lbuf_append(&lbuf, "  -V, --version                 %s\n",
+	_("display version information and exit"));
+    sudo_lbuf_append(&lbuf, "  -v, --validate                %s\n",
+	_("update user's timestamp without running a command"));
+    sudo_lbuf_append(&lbuf, "  --                            %s\n",
+	_("stop processing command line arguments"));
+    sudo_lbuf_print(&lbuf);
+    sudo_lbuf_destroy(&lbuf);
     sudo_debug_exit_int(__func__, __FILE__, __LINE__, sudo_debug_subsys, 0);
     exit(0);
 }
