@@ -190,7 +190,7 @@ sudo_openat(int dfd, const char *path, int flags, mode_t mode)
 static int
 sudo_edit_openat_nofollow(int dfd, char *path, int oflags, mode_t mode)
 {
-    debug_decl(sudo_edit_open_nofollow, SUDO_DEBUG_EDIT)
+    debug_decl(sudo_edit_openat_nofollow, SUDO_DEBUG_EDIT)
 
     debug_return_int(openat(dfd, path, oflags|O_NOFOLLOW, mode));
 }
@@ -226,20 +226,51 @@ sudo_edit_is_symlink(int fd, char *path)
 static int
 sudo_edit_openat_nofollow(int dfd, char *path, int oflags, mode_t mode)
 {
-    struct stat sb1, sb2;
-    int fd;
+    int fd = -1, odfd = -1;
+    struct stat sb;
     debug_decl(sudo_edit_openat_nofollow, SUDO_DEBUG_EDIT)
 
-    fd = openat(dfd, path, oflags, mode);
-    if (fd == -1)
+    /* Save cwd and chdir to dfd */
+    if ((odfd = open(".", O_RDONLY)) == -1)
 	debug_return_int(-1);
+    if (fchdir(dfd) == -1) {
+	close(odfd);
+	debug_return_int(-1);
+    }
 
+    /*
+     * Check if path is a symlink.  This is racey but we detect whether
+     * we lost the race in sudo_edit_is_symlink() after the open.
+     */
+    if (lstat(path, &sb) == -1 && errno != ENOENT)
+	goto done;
+    if (S_ISLNK(sb.st_mode)) {
+	errno = ELOOP;
+	goto done;
+    }
+
+    fd = open(path, oflags, mode);
+    if (fd == -1)
+	goto done;
+
+    /*
+     * Post-open symlink check.  This will leave a zero-length file if
+     * O_CREAT was specified but it is too dangerous to try and remove it.
+     */
     if (sudo_edit_is_symlink(fd, path)) {
 	close(fd);
 	fd = -1;
 	errno = ELOOP;
     }
 
+done:
+    /* Restore cwd */
+    if (odfd != -1) {
+	if (fchdir(odfd) == -1)
+	    sudo_fatal(_("unable to restore current working directory"));
+	close(odfd);
+    }
+
     debug_return_int(fd);
 }
 #endif /* O_NOFOLLOW */
@@ -384,8 +415,10 @@ sudo_edit_open_nonwritable(char *path, int oflags, mode_t mode,
 	 * writable directories.
 	 */
 	is_writable = dir_is_writable(dfd, &user_details, command_details);
-	if (is_writable == -1)
+	if (is_writable == -1) {
+	    close(dfd);
 	    debug_return_int(-1);
+	}
 
 	while (path[0] == '/')
 	    path++;
@@ -448,10 +481,23 @@ sudo_edit_open(char *path, int oflags, mode_t mode,
     struct command_details *command_details)
 {
     const int sflags = command_details ? command_details->flags : 0;
-    struct stat sb1, sb2;
+    struct stat sb;
     int fd;
     debug_decl(sudo_edit_open, SUDO_DEBUG_EDIT)
 
+    /*
+     * Check if path is a symlink.  This is racey but we detect whether
+     * we lost the race in sudo_edit_is_symlink() after the file is opened.
+     */
+    if (!ISSET(sflags, CD_SUDOEDIT_FOLLOW)) {
+	if (lstat(path, &sb) == -1 && errno != ENOENT)
+	    debug_return_int(-1);
+	if (S_ISLNK(sb.st_mode)) {
+	    errno = ELOOP;
+	    debug_return_int(-1);
+	}
+    }
+
     if (ISSET(sflags, CD_SUDOEDIT_CHECKDIR) && user_details.uid != ROOT_UID) {
 	fd = sudo_edit_open_nonwritable(path, oflags|O_NONBLOCK, mode,
 	    command_details);
@@ -463,6 +509,10 @@ sudo_edit_open(char *path, int oflags, mode_t mode,
     if (!ISSET(oflags, O_NONBLOCK))
 	(void) fcntl(fd, F_SETFL, fcntl(fd, F_GETFL, 0) & ~O_NONBLOCK);
 
+    /*
+     * Post-open symlink check.  This will leave a zero-length file if
+     * O_CREAT was specified but it is too dangerous to try and remove it.
+     */
     if (!ISSET(sflags, CD_SUDOEDIT_FOLLOW) && sudo_edit_is_symlink(fd, path)) {
 	close(fd);
 	fd = -1;
@@ -498,7 +548,8 @@ sudo_edit_create_tfiles(struct command_details *command_details,
 	rc = -1;
 	switch_user(command_details->euid, command_details->egid,
 	    command_details->ngroups, command_details->groups);
-	ofd = sudo_edit_open(files[i], O_RDONLY, 0644, command_details);
+	ofd = sudo_edit_open(files[i], O_RDONLY,
+	    S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH, command_details);
 	if (ofd != -1 || errno == ENOENT) {
 	    if (ofd == -1) {
 		/* New file, verify parent dir exists unless in cwd. */
@@ -556,20 +607,27 @@ sudo_edit_create_tfiles(struct command_details *command_details,
 	    sudo_fatal("seteuid(ROOT_UID)");
 	if (tfd == -1) {
 	    sudo_warn("mkstemps");
+	    if (ofd != -1)
+		close(ofd);
 	    debug_return_int(-1);
 	}
 	if (ofd != -1) {
-	    while ((nread = read(ofd, buf, sizeof(buf))) != 0) {
+	    while ((nread = read(ofd, buf, sizeof(buf))) > 0) {
 		if ((nwritten = write(tfd, buf, nread)) != nread) {
 		    if (nwritten == -1)
 			sudo_warn("%s", tf[j].tfile);
 		    else
 			sudo_warnx(U_("%s: short write"), tf[j].tfile);
-		    close(ofd);
-		    close(tfd);
-		    debug_return_int(-1);
+		    break;
 		}
 	    }
+	    if (nread != 0) {
+		if (nread < 0)
+		    sudo_warn("%s", files[i]);
+		close(ofd);
+		close(tfd);
+		debug_return_int(-1);
+	    }
 	    close(ofd);
 	}
 	/*
@@ -616,7 +674,8 @@ sudo_edit_copy_tfiles(struct command_details *command_details,
 	    "seteuid(%u)", user_details.uid);
 	if (seteuid(user_details.uid) != 0)
 	    sudo_fatal("seteuid(%d)", (int)user_details.uid);
-	tfd = sudo_edit_open(tf[i].tfile, O_RDONLY, 0644, NULL);
+	tfd = sudo_edit_open(tf[i].tfile, O_RDONLY,
+	    S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH, NULL);
 	if (tfd != -1)
 	    rc = fstat(tfd, &sb);
 	sudo_debug_printf(SUDO_DEBUG_INFO|SUDO_DEBUG_LINENO,
@@ -650,8 +709,8 @@ sudo_edit_copy_tfiles(struct command_details *command_details,
 	}
 	switch_user(command_details->euid, command_details->egid,
 	    command_details->ngroups, command_details->groups);
-	ofd = sudo_edit_open(tf[i].ofile, O_WRONLY|O_TRUNC|O_CREAT, 0644,
-	    command_details);
+	ofd = sudo_edit_open(tf[i].ofile, O_WRONLY|O_TRUNC|O_CREAT,
+	    S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH, command_details);
 	switch_user(ROOT_UID, user_details.egid,
 	    user_details.ngroups, user_details.groups);
 	if (ofd == -1) {
@@ -788,7 +847,7 @@ selinux_edit_copy_tfiles(struct command_details *command_details,
     struct tempfile *tf, int nfiles, struct timespec *times)
 {
     char **sesh_args, **sesh_ap;
-    int i, rc, sesh_nargs, rval = 1;
+    int i, rc, sesh_nargs, ret = 1;
     struct command_details saved_command_details;
     struct timespec ts;
     struct stat sb;
@@ -848,7 +907,7 @@ selinux_edit_copy_tfiles(struct command_details *command_details,
 	rc = run_command(command_details);
 	switch (rc) {
 	case SESH_SUCCESS:
-	    rval = 0;
+	    ret = 0;
 	    break;
 	case SESH_ERR_NO_FILES:
 	    sudo_warnx(_("unable to copy temporary files back to their original location"));
@@ -863,13 +922,14 @@ selinux_edit_copy_tfiles(struct command_details *command_details,
 	    break;
 	}
     }
+    free(sesh_args);
 
     /* Restore saved command_details. */
     command_details->command = saved_command_details.command;
     command_details->flags = saved_command_details.flags;
     command_details->argv = saved_command_details.argv;
 
-    debug_return_int(rval);
+    debug_return_int(ret);
 }
 #endif /* HAVE_SELINUX */
 
@@ -882,7 +942,7 @@ sudo_edit(struct command_details *command_details)
 {
     struct command_details saved_command_details;
     char **nargv = NULL, **ap, **files = NULL;
-    int errors, i, ac, nargc, rval;
+    int errors, i, ac, nargc, rc;
     int editor_argc = 0, nfiles = 0;
     struct timespec times[2];
     struct tempfile *tf = NULL;
@@ -967,7 +1027,7 @@ sudo_edit(struct command_details *command_details)
     command_details->ngroups = user_details.ngroups;
     command_details->groups = user_details.groups;
     command_details->argv = nargv;
-    rval = run_command(command_details);
+    rc = run_command(command_details);
     if (sudo_gettime_real(&times[1]) == -1) {
 	sudo_warn(U_("unable to read the clock"));
 	goto cleanup;
@@ -994,7 +1054,7 @@ sudo_edit(struct command_details *command_details)
 	free(tf[i].tfile);
     free(tf);
     free(nargv);
-    debug_return_int(errors ? 1 : rval);
+    debug_return_int(errors ? 1 : rc);
 
 cleanup:
     /* Clean up temp files and return. */
